from amaranth import Module, Cat, Mux, ShapeLike, Signal, Assert, Array
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out
from amaranth.lib.memory import Memory, ReadPort, WritePort
from amaranth.utils import ceil_log2

from .common import Rgb666Layout, Hub75Stream, Hub75Ctrl, Hub75Data, Rgb888Layout


class SwapBuffer(wiring.Component):
    """A pair of BRAMs for holdling line data that are swapped between using an external signal.

    Contains one write port, and one read port. They are attached to separate BRAMs to allow for
    one to be read at the same time another is being updated.

    Parameters:
        depth: i32, depth of each memory buffer
        shape: ShapeLike, the underlying shape of the memory.

    Signals:
        write_port: WritePort.Signature(width, shape)
        read_port: ReadPort.Signature(width, shape)
        selector: In(1)

    """

    def __init__(self, shape: ShapeLike, depth: int):
        super().__init__(
            {
                "selector": In(1),
                "read_port": In(
                    ReadPort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
                "write_port": In(
                    WritePort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
            }
        )
        self.data_shape = shape
        self.depth = depth

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        m.submodules.bram0 = self.bram0 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )
        m.submodules.bram1 = self.bram1 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )

        rd0 = self.bram0.read_port()
        wr0 = self.bram0.write_port()
        rd1 = self.bram1.read_port()
        wr1 = self.bram1.write_port()

        m.d.comb += [
            # wr addres
            wr0.addr.eq(self.write_port.addr),
            wr1.addr.eq(self.write_port.addr),
            # write enables are based on selector
            wr0.en.eq(~self.selector),
            wr1.en.eq(self.selector),
            # connect write data. This is FINE because
            # there is one driver (the external writer)
            # and we en based on selector so the other one isn't active
            wr0.data.eq(self.write_port.data),
            wr1.data.eq(self.write_port.data),
            # connect rd address lines
            rd0.addr.eq(self.read_port.addr),
            rd1.addr.eq(self.read_port.addr),
            rd0.en.eq(~self.selector),
            rd1.en.eq(self.selector),
            # we do this because the read_data lines are driven, this prevents
            # double-driver situations even though we en using selector above
            self.read_port.data.eq(Mux(self.selector, rd1.data, rd0.data)),
        ]

        return m


class Hub75DataDriver(wiring.Component):
    def __init__(
        self,
        panel_length: int = 128,
        data_shape=Rgb888Layout,
        double_fetch=True,
        *,
        src_loc_at=0,
    ):
        self.panel_length = panel_length
        self.double_fetch = double_fetch
        # if we're using double_fetch then each bram cell is exactly one color,
        # if we're doing single fetch, it's 2 colors stacked.
        bram_shape = data_shape if double_fetch else data.ArrayLayout(data_shape, 2)
        addr_width = (
            ceil_log2(panel_length * 2) if double_fetch else ceil_log2(panel_length)
        )
        super().__init__(
            {
                "bcm_select": In(ceil_log2(data_shape["red"].shape.width)),
                "done": Out(1),
                "start": In(1),
                "bram": In(
                    ReadPort.Signature(
                        addr_width=addr_width,
                        shape=bram_shape,
                    )
                ),
                "data": Out(Hub75Data()),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)

        pixnum = Signal(range(self.panel_length), init=self.panel_length - 1)
        if self.double_fetch:
            pixrow = Signal(1)
            m.d.comb += self.bram.addr.eq(Cat(pixrow, pixnum))
            ram_rgb_slice = Cat(
                self.bram.data["red"].bit_select(self.bcm_select, 1),
                self.bram.data["blue"].bit_select(self.bcm_select, 1),
                self.bram.data["green"].bit_select(self.bcm_select, 1),
            )
        else:
            ram_rgb_slice = Array(
                [
                    Cat(
                        self.bram.data[0]["red"].bit_select(self.bcm_select, 1),
                        self.bram.data[0]["blue"].bit_select(self.bcm_select, 1),
                        self.bram.data[0]["green"].bit_select(self.bcm_select, 1),
                    ),
                    Cat(
                        self.bram.data[1]["red"].bit_select(self.bcm_select, 1),
                        self.bram.data[1]["blue"].bit_select(self.bcm_select, 1),
                        self.bram.data[1]["green"].bit_select(self.bcm_select, 1),
                    ),
                ]
            )

            m.d.comb += self.bram.addr.eq(pixnum)

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    self.done.eq(0),
                    counter.eq(0),
                    pixnum.eq(pixnum.reset),
                ]
                if self.double_fetch:
                    m.d.sync += pixrow.eq(0)
                with m.If(self.start == 1):
                    m.d.sync += self.bram.en.eq(1)
                    m.next = "prefetch" if self.double_fetch else "writerow"
            with m.State("prefetch"):
                # TODO: do we need this
                m.next = "writerow"

            with m.State("writerow"):
                if self.double_fetch:
                    c = counter[0:1]
                    with m.If(c == 0):
                        m.d.sync += [
                            self.data.rgb0.eq(ram_rgb_slice),
                            pixrow.eq(1),
                        ]
                    with m.If(c == 1):
                        m.d.sync += self.data.rgb1.eq(ram_rgb_slice)
                    with m.If(c == 2):
                        pass

                    with m.If(c == 3):
                        m.d.sync += [
                            counter.eq(0),
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                        ]
                        with m.If(pixnum == 0):
                            m.next = "done"
                else:
                    with m.If(counter[0] == 0):
                        m.d.sync += [
                            self.data.rgb0.eq(ram_rgb_slice[0]),
                            self.data.rgb1.eq(ram_rgb_slice[1]),
                        ]
                    with m.If(counter[0] == 1):
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            counter.eq(0),
                        ]
                        with m.If(pixnum == 0):
                            m.next = "done"
            with m.State("done"):
                m.d.sync += [self.done.eq(1), self.bram.en.eq(0)]
                m.next = "init"

        return m


class Hub75StringDriver(wiring.Component):
    """A data driver for Hub75 panels. This accesses the line memory and feeds out the data.
    It is controlled by a Hub75Coordinator to signal when it should run and what bit of the data
    it should send.
    """

    def __init__(self, panel_length: int = 128, *, src_loc_at=0):
        self.panel_length = panel_length
        super().__init__(
            {
                "bcm_select": In(3),
                "done": Out(1),
                "start": In(1),
                "bram_port": In(
                    ReadPort.Signature(
                        addr_width=ceil_log2(panel_length),
                        shape=data.ArrayLayout(Rgb666Layout, 2),
                    )
                ),
                "display_out": Out(Hub75Data()),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        self._counter = counter = Signal(32)  # unused count is optimized out
        m.d.sync += counter.eq(counter + 1)

        ram_rgb0_slice = Cat(
            self.bram_port.data[0]["red"].bit_select(self.bcm_select, 1),
            self.bram_port.data[0]["blue"].bit_select(self.bcm_select, 1),
            self.bram_port.data[0]["green"].bit_select(self.bcm_select, 1),
        )
        ram_rgb1_slice = Cat(
            self.bram_port.data[1]["red"].bit_select(self.bcm_select, 1),
            self.bram_port.data[1]["blue"].bit_select(self.bcm_select, 1),
            self.bram_port.data[1]["green"].bit_select(self.bcm_select, 1),
        )
        m.d.comb += [
            self.display_out.rgb0.eq(ram_rgb0_slice),
            self.display_out.rgb1.eq(ram_rgb1_slice),
        ]
        m.d.sync += Assert(self.bcm_select < 6)

        pixnum = Signal(range(self.panel_length), init=self.panel_length - 1)
        m.d.comb += self.bram_port.addr.eq(pixnum)

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    self.done.eq(0),
                    counter.eq(0),
                    pixnum.eq(self.panel_length - 1),
                ]
                with m.If(self.start == 1):
                    m.d.sync += self.bram_port.en.eq(1)
                    m.next = "writerow"
            with m.State("writerow"):
                with m.If(counter[0] == 0):
                    # do nothing
                    pass

                with m.If((counter[0] == 1) & (pixnum != 0)):
                    m.d.sync += pixnum.eq(pixnum - 1)

                with m.Else():
                    m.next = "done"
            with m.State("done"):
                m.d.sync += [self.done.eq(1), self.bram_port.en.eq(0)]
                m.next = "init"

        return m


class Hub75Coordinator(wiring.Component):
    """A shared-control hub75 driver"""

    def __init__(self, n_strings=1):
        self.n_strings = n_strings
        super().__init__(
            {
                "hub75": Out(Hub75Ctrl(n_strings)),
                # TODO: fetching routine? maybe it's passed through.
            }
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()
        # swapline is which buffer we are using vs sending out.
        swapline = Signal(1)

        # for each string, spawn a swapbuffer + stringdriver and connect.
        # don't worry about fetching for now.
        self.strings = strings = []
        self.buffers = bufs = []
        donearr = []
        startStrings = Signal(1)
        stringsDone = Signal(1)

        for i in range(self.n_strings):
            sb = SwapBuffer(depth=128, shape=data.ArrayLayout(Rgb666Layout, 2))
            bufs.append(sb)
            stringdriver = Hub75StringDriver(128)
            strings.append(stringdriver)
            wiring.connect(m, sb.read_port, stringdriver.bram_port)
            m.d.comb += [
                self.data[i].eq(stringdriver.display_out),
                stringdriver.start.eq(startStrings),
                sb.selector.eq(swapline),
            ]
            m.submodules += [sb, stringdriver]
            donearr.append(stringdriver.done)
        # combine the done signals into one signal with AND-reduction
        m.d.comb += stringsDone.eq(Cat(*donearr).all())

        self.addr = Signal(5)
        # handle the fetch side.
        # WIP: pass in fetcher/pixgen/geometry.
        # right now we assume that it's just one panel,
        # address is (string_number, hi/lo, addr)

        for i in range(self.n_strings):
            lookup_addr = Cat(i, self.addr, 0)
            # generate a sequence of transfers.

        with m.FSM():
            with m.State("init"):
                # go to preload.
                pass

            with m.State("preload"):
                pass

            with m.State("run"):
                pass

            with m.State("swap"):
                pass

        return m


# fetch line
#


class Hub75EDriver(wiring.Component):
    """An optimized driver for hub75 strings.
    This version is faster than most implementations by merging the exposure
    period and the data-write period to happen simultaneously. As a result,
    the display can be brighter due to higher duty cycle.


    NOTICE: this is a direct port of the old verilog code. It isn't up to date with the
    modified structure. Notably, it uses RGB888 with double-fetch (4xclocking)
    """

    start: In(1)
    done: Out(1)
    out: Out(Hub75Stream())
    buf_addr: Out(9)
    buf_data: In(36)

    row_depth = 128
    bit_depth = 8
    bcm_len = 32

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        bcm_shift: Signal = Signal(4, init=7)
        m.d.sync += counter.eq(counter + 1)
        # construct helper signals.
        ram_r = self.buf_data[16:23]
        ram_b = self.buf_data[8:15]
        ram_g = self.buf_data[0:7]

        ram_rgb_slice = Cat(
            ram_r.bit_select(bcm_shift, 1),
            ram_g.bit_select(bcm_shift, 1),
            ram_b.bit_select(bcm_shift, 1),
        )
        pixnum = Signal(8, reset=127)
        pixrow = Signal(1, reset=0)
        m.d.comb += self.buf_addr.eq(Cat(pixrow, pixnum))

        should_clock = counter < (128 * 2 + 1)
        should_expose = (counter < (32 << (bcm_shift + 1) + 1)).bool() & (
            bcm_shift != 7
        ).bool()

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    bcm_shift.eq(7),
                    counter.eq(0),
                    self.done.eq(0),
                    pixnum.eq(127),
                    pixrow.eq(0),
                ]
                with m.If(self.start == 1):
                    m.next = "prefetch"
            with m.State("prefetch"):
                with m.If(counter == 0):
                    m.d.sync += pixrow.eq(0)
                with m.Elif(counter == 1):
                    m.d.sync += pixrow.eq(1)
                with m.Elif(counter == 2):
                    m.d.sync += self.out.rgb0.eq(ram_rgb_slice)
                with m.Elif(counter == 3):
                    m.d.sync += [self.out.rgb1.eq(ram_rgb_slice), counter.eq(0)]
                    m.next = "writerow"
            with m.State("writerow"):
                # expose if we haven't done it for long enough yet.
                m.d.sync += self.out.oe.eq(~(should_expose))
                with m.If(should_clock):
                    # clock is high on entry
                    m.d.sync += self.out.display_clk.eq(counter[1] == 0)

                    with m.If(counter[0:1] == 0):
                        # rising edge of the clock
                        m.d.sync += pixrow.eq(0)

                    with m.If(counter[0:1] == 1):
                        m.d.sync += [
                            pixrow.eq(1),
                            self.out.rgb0.eq(ram_rgb_slice),
                        ]
                    with m.If(counter[0:1] == 2):
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                            self.out.rgb1.eq(ram_rgb_slice),
                        ]

                with m.Elif(~(should_expose)):
                    # we are done both feeding in the new data and exposing the previous.
                    m.d.sync += [counter.eq(0), self.out.display_clk.eq(0)]
                    m.next = "latchout"
            with m.State("latchout"):
                m.d.sync += [
                    pixnum.eq(127),
                    self.out.latch.eq(1),
                ]
                with m.If(counter > 3):
                    m.d.sync += self.out.latch.eq(0)
                    m.d.sync += counter.eq(0)
                    with m.If(bcm_shift == 0):
                        m.next = "finish"
                    with m.Else():
                        m.d.sync += bcm_shift.eq(bcm_shift - 1)
                        m.next = "prefetch"

            with m.State("finish"):
                m.d.sync += Assert(bcm_shift == 0, "finish without bcm shift 0")

                with m.If(counter < (32)):
                    m.d.sync += self.out.oe.eq(0)
                with m.Else():
                    m.d.sync += [self.out.oe.eq(1), self.done.eq(1)]
                    m.next = "init"

        return m


if __name__ == "__main__":
    m = Hub75EDriver()
    from amaranth.cli import main

    main(m)
